@@ -5,184 +5,207 @@ import json
 import types
 import re
 from functools import partial
-from wsgiref.headers import Headers as BaseHeaders
 import itertools
 import logging
 import inspect
 import copy
 from socket import gethostname
+import cgi
+import io
 
-from datatypes import Url as BaseUrl, Host
+from datatypes import Url as BaseUrl, Host, Headers as BaseHeaders, Environ
 
 from .compat import *
-from .decorators.utils import _property
-from .utils import AcceptHeader, ByteString, MimeType, String, Base64, Deepcopy
+from .decorators.utils import property # must be .utils because circular dep
+from .utils import (
+    AcceptHeader,
+    MimeType,
+    Base64,
+    Deepcopy,
+    FileWrapper,
+)
 
 
 logger = logging.getLogger(__name__)
 
 
-class Headers(BaseHeaders, Mapping):
-    """handles headers, see wsgiref.Headers link for method and use information
+class Headers(BaseHeaders):
+    def is_plain(self):
+        """return True if body's content-type is text/plain"""
+        ct = self.get("Content-Type", "")
+        return "plain" in ct
 
-    Handles normalizing of header names, the problem with headers is they can
-    be in many different forms and cases and stuff (eg, CONTENT_TYPE and Content-Type),
-    so this handles normalizing the header names so you can request Content-Type
-    or CONTENT_TYPE and get the same value.
+    def is_json(self):
+        """return True if body's content-type is application/json"""
+        ct = self.get("Content-Type", "")
+        return "json" in ct
 
-    This has the same interface as Python's built-in wsgiref.Headers class but
-    makes it even more dict-like and will return titled header names when iterated
-    or anything (eg, Content-Type instead of all lowercase content-type)
+    def is_urlencoded(self):
+        """return True if body's content-type is application/x-www-form-urlencoded"""
+        ct = self.get("Content-Type", "")
+        return ("form-urlencoded" in ct) or ("form-data" in ct)
 
-    http headers spec:
-        https://www.w3.org/Protocols/rfc2616/rfc2616-sec14.html
+    def is_multipart(self):
+        """return True if body's content-type is multipart/form-data"""
+        ct = self.get("Content-Type", "")
+        return "multipart" in ct
 
-    wsgiref class docs:
-        https://docs.python.org/2/library/wsgiref.html#module-wsgiref.headers
-        https://hg.python.org/cpython/file/2.7/Lib/wsgiref/headers.py
-    actual python3 code:
-        https://github.com/python/cpython/blob/master/Lib/wsgiref/headers.py
+
+class Body(cgi.FieldStorage, object):
+    """Wraps the default FieldStorage to handle json and also recovers when the
+    input fails to parse correctly
+
+    https://github.com/python/cpython/blob/2.7/Lib/cgi.py
+    https://github.com/python/cpython/blob/3.8/Lib/cgi.py
     """
-    def __init__(self, headers=None, **kwargs):
-        super(Headers, self).__init__([])
-        self.update(headers, **kwargs)
+    FieldStorageClass = cgi.FieldStorage
 
-    def _convert_string_part(self, bit):
-        """each part of a header will go through this method, this allows further
-        normalization of each part, so a header like FOO_BAR would call this method
-        twice, once with foo and again with bar
+    @property(cached="_args")
+    def args(self):
+        return getattr(self, "json_args", [])
 
-        :param bit: string, a part of a header all lowercase
-        :returns: string, the normalized bit
-        """
-        if bit == "websocket":
-            bit = "WebSocket"
-        else:
-            bit = bit.title()
-        return bit
+    @property(cached="_kwargs")
+    def kwargs(self):
+        body_kwargs = {}
+        body_kwargs.update(getattr(self, "json_kwargs", {}))
 
-    def _convert_string_name(self, k):
-        """converts things like FOO_BAR to Foo-Bar which is the normal form"""
-        k = String(k, "iso-8859-1")
-        bits = k.lower().replace('_', '-').split('-')
-        return "-".join((self._convert_string_part(bit) for bit in bits))
+        # we only have a list when we had a multiport or data-form submission
+        if getattr(self, "list"):
+            for field_name in self.keys():
+                body_field = self[field_name]
+                if body_field.filename:
+                    body_kwargs[field_name] = FileWrapper(
+                        body_field.file,
+                        name=body_field.filename,
+                        filename=body_field.filename,
+                        type=body_field.type,
+                        raw=body_field,
+                    )
 
-    def _convert_string_type(self, v):
-        """Override the internal method wsgiref.headers.Headers uses to check values
-        to make sure they are strings"""
-        # wsgiref.headers.Headers expects a str() (py3) or unicode (py2), it
-        # does not accept even a child of str, so we need to convert the String
-        # instance to the actual str, as does the python wsgi methods, so even
-        # though we override this method we still return raw() strings so we get
-        # passed all the type(v) == "str" checks
-        # sadly, this method is missing in 2.7
-        # https://github.com/python/cpython/blob/2.7/Lib/wsgiref/headers.py
-        return String(v).raw()
+                else:
+                    body_kwargs[field_name] = body_field.value
 
-    def get_all(self, name):
-        name = self._convert_string_name(name)
-        return super(Headers, self).get_all(name)
 
-    def get(self, name, default=None):
-        name = self._convert_string_name(name)
-        return super(Headers, self).get(name, default)
+        return body_kwargs
 
-    def __delitem__(self, name):
-        name = self._convert_string_name(name)
-        return super(Headers, self).__delitem__(name)
+    def __init__(self, fp, request, **kwargs):
+        if request.headers.get('transfer-encoding', "").lower().startswith("chunked"):
+            raise IOError("Chunked bodies are not supported")
 
-    def __setitem__(self, name, val):
-        name = self._convert_string_name(name)
-        if is_py2:
-            val = self._convert_string_type(val)
-        return super(Headers, self).__setitem__(name, val)
+        self.request = request
 
-    def setdefault(self, name, val):
-        name = self._convert_string_name(name)
-        if is_py2:
-            val = self._convert_string_type(val)
-        return super(Headers, self).setdefault(name, val)
+        # py3 compatibility
+        self.encoding = request.encoding
+        self.errors = "replace"
+        self.max_num_fields = None
 
-    def add_header(self, name, val, **params):
-        name = self._convert_string_name(name)
-        if is_py2:
-            val = self._convert_string_type(val)
-        return super(Headers, self).add_header(name, val, **params)
+        self.name = self.filename = self.value = None
+        self.length = int(request.headers.get("CONTENT_LENGTH", -1))
+        self.fp = fp
+        self.list = None
 
-    def keys(self):
-        return [k for k, v in self._headers]
+        if self.length > 0:
+            if request.is_json():
+                self.read_json()
 
-    def items(self):
-        for k, v in self._headers:
-            yield k, v
-
-    def iteritems(self):
-        return self.items()
-
-    def iterkeys(self):
-        for k in self.keys():
-            yield k
-
-    def __iter__(self):
-        for k, v in self._headers:
-            yield k
-
-    def pop(self, name, *args, **kwargs):
-        """remove and return the value at name if it is in the dict
-
-        This uses *args and **kwargs instead of default because this will raise
-        a KeyError if default is not supplied, and if it had a definition like
-        (name, default=None) you wouldn't be able to know if default was provided
-        or not
-
-        :param name: string, the key we're looking for
-        :param default: mixed, the value that would be returned if name is not in
-            dict
-        :returns: the value at name if it's there
-        """
-        val = self.get(name)
-        if val is None:
-            if args:
-                val = args[0]
-            elif "default" in kwargs:
-                val = kwargs["default"]
             else:
-                raise KeyError(name)
+                kwargs.setdefault("keep_blank_values", True)
 
-        else:
-            del self[name]
+                # so FieldStorage parses the body in the constructor and if it fails
+                # then the body instance won't be created, so this is set to True and
+                # the error is handled in read_urlencoded
+                kwargs["strict_parsing"] = True
 
-        return val
-
-    def update(self, headers, **kwargs):
-        if not headers: headers = {}
-        if isinstance(headers, Mapping):
-            headers.update(kwargs)
-            headers = headers.items()
-
-        else:
-            if kwargs:
-                headers = itertools.chain(
-                    headers,
-                    kwargs.items()
+                super(Body, self).__init__(
+                    fp=fp,
+                    headers=request.headers,
+                    environ=request.environ,
+                    **kwargs
                 )
 
-        for k, v in headers:
-            self[k] = v
+    def is_plain(self):
+        """return True if body's content-type is text/plain"""
+        return self.request.headers.is_plain()
 
-    def copy(self):
-        return Deepcopy().copy(self)
+    def is_json(self):
+        """return True if body's content-type is application/json"""
+        return self.request.headers.is_json()
 
-    def list(self):
-        """Return all the headers as a list of headers instead of a dict"""
-        return [": ".join(h) for h in self.items() if h[1]]
+    def is_urlencoded(self):
+        """return True if body's content-type is application/x-www-form-urlencoded"""
+        return self.request.headers.is_urlencoded()
 
+    def is_multipart(self):
+        """return True if body's content-type is multipart/form-data"""
+        return self.request.headers.is_multipart()
 
-class Environ(Headers):
-    """just like Headers but allows any values (headers converts everything to unicode
-    string)"""
-    def _convert_string_type(self, v):
-        return v
+    def read_json(self):
+        body = self.fp.read(self.length)
+        self.file = io.BytesIO(body)
+
+        body_args = []
+        body_kwargs = {}
+        b = json.loads(body)
+        if isinstance(b, list):
+            body_args = b
+
+        elif isinstance(b, dict):
+            body_kwargs = b
+
+        else:
+            body_args = [b]
+
+        self.json_args = body_args
+        self.json_kwargs = body_kwargs
+
+    def read_urlencoded(self):
+        """Internal: read data in query string format."""
+        body = self.fp.read(self.length)
+        self.file = io.BytesIO(body)
+
+        qs = String(body, self.encoding, self.errors)
+
+        if self.qs_on_post:
+            qs += '&' + self.qs_on_post
+
+        try:
+            if is_py2:
+                query = parse.parse_qsl(
+                    qs,
+                    self.keep_blank_values,
+                    self.strict_parsing,
+                )
+
+            else:
+                query = parse.parse_qsl(
+                    qs,
+                    self.keep_blank_values,
+                    self.strict_parsing,
+                    encoding=self.encoding,
+                    errors=self.errors,
+                    max_num_fields=self.max_num_fields
+                )
+
+        except ValueError:
+            # if the right headers were sent then this should error
+            if self.is_urlencoded() or self.is_multipart():
+                raise
+
+        else:
+            self.list = [cgi.MiniFieldStorage(key, value) for key, value in query]
+            self.skip_lines()
+
+    def make_file(self, *args, **kwargs):
+        return io.BytesIO()
+
+    def seek(self, *args, **kwargs):
+        return self.file.seek(*args, **kwargs)
+
+    def read(self, *args, **kwargs):
+        return self.file.read(*args, **kwargs)
+
+    def tell(self, *args, **kwargs):
+        return self.file.tell(*args, **kwargs)
 
 
 class Url(BaseUrl):
@@ -258,6 +281,8 @@ class Url(BaseUrl):
 
 
 class Http(object):
+    header_class = Headers
+
     def __init__(self):
         self.headers = Headers()
 
@@ -326,14 +351,13 @@ class Http(object):
     def __deepcopy__(self, memodict=None):
         memodict = memodict or {}
 
-        if self.controller_info:
-            memodict.setdefault("controller_info", self.controller_info)
+        memodict.setdefault("controller_info", getattr(self, "controller_info", {}))
+        memodict.setdefault("body", getattr(self, "body", None))
 
         return Deepcopy(ignore_private=True).copy(self, memodict)
 
     def is_json(self):
-        ct = self.get_header('Content-Type')
-        return ct.lower().rfind("json") >= 0 if ct else False
+        return self.headers.is_json()
 
 
 class Request(Http):
@@ -353,7 +377,6 @@ class Request(Http):
     query -- the ?name=val portion of a url
     query_kwargs -- tied to query, the values in query but converted to a dict {name: val}
     '''
-
     environ = None
     """holds all the values that aren't considered headers but usually get passed with the request"""
 
@@ -366,6 +389,9 @@ class Request(Http):
     controller_info = None
     """will hold the controller information for the request, populated from the Call"""
 
+    body_class = Body
+    """see create_body()"""
+
     @property
     def accept_encoding(self):
         """The encoding the client requested the response to use"""
@@ -378,7 +404,7 @@ class Request(Http):
             ret = bits[0]
         return ret
 
-    @_property
+    @property(cached="_encoding")
     def encoding(self):
         """the character encoding of the request, usually only set in POST type requests"""
         encoding = None
@@ -422,7 +448,7 @@ class Request(Http):
 
         return client_id, client_secret
 
-    @_property(read_only=True)
+    @property(read_only="_ips")
     def ips(self):
         """return all the possible ips of this request, this will include public and private ips"""
         r = []
@@ -441,7 +467,7 @@ class Request(Http):
 
         return r
 
-    @_property(read_only=True)
+    @property(read_only="_ip")
     def ip(self):
         """return the public ip address"""
         r = ''
@@ -476,18 +502,18 @@ class Request(Http):
 
         return r
 
-    @_property
+    @property(cached="_host")
     def host(self):
         """return the request host"""
         return self.get_header("host")
 
-    @_property
+    @property(cached="_scheme")
     def scheme(self):
         """return the request scheme (eg, http, https)"""
         scheme = self.environ.get('wsgi.url_scheme', "http")
         return scheme
 
-    @_property
+    @property(cached="_port")
     def port(self):
         """return the server port"""
         return int(self.environ.get('SERVER_PORT', 0))
@@ -528,7 +554,7 @@ class Request(Http):
         )
         return u
 
-    @_property
+    @property(cached="_path")
     def path(self):
         """path part of a url (eg, http://host.com/path?query=string)"""
         self._path = ''
@@ -536,7 +562,7 @@ class Request(Http):
         path = "/{}".format("/".join(path_args))
         return path
 
-    @_property
+    @property(cached="_path_args")
     def path_args(self):
         """the path converted to list (eg /foo/bar becomes [foo, bar])"""
         self._path_args = []
@@ -544,7 +570,7 @@ class Request(Http):
         path_args = list(filter(None, path.split('/')))
         return path_args
 
-    @_property
+    @property(cached="_query")
     def query(self):
         """query_string part of a url (eg, http://host.com/path?query=string)"""
         self._query = query = ""
@@ -553,7 +579,7 @@ class Request(Http):
         if query_kwargs: query = urlencode(query_kwargs, doseq=True)
         return query
 
-    @_property
+    @property(cached="_query_kwargs")
     def query_kwargs(self):
         """{foo: bar, baz: che}"""
         self._query_kwargs = query_kwargs = {}
@@ -575,6 +601,9 @@ class Request(Http):
         self.body_kwargs = {}
         super(Request, self).__init__()
 
+    def create_body(self, body):
+        return self.body_class(body, self)
+
     def version(self, content_type="*/*"):
         """
         versioning is based off of this post 
