@@ -14,6 +14,59 @@ from endpoints.exception import CallError
 
 
 class ControllerTest(TestCase):
+    def test_any_1(self):
+        c = Server(contents=[
+            "from endpoints import Controller, param",
+            "class Default(Controller):",
+            "    def ANY(self):",
+            "        return 'any'",
+        ])
+
+        res = c.handle("/")
+        self.assertEqual(200, res.code)
+        self.assertEqual("any", res.body)
+
+        res = c.handle("/", method="POST")
+        self.assertEqual(200, res.code)
+        self.assertEqual("any", res.body)
+
+        res = c.handle("/foo/bar")
+        self.assertEqual(404, res.code)
+
+    def test_any_2(self):
+        c = Server(contents=[
+            "from endpoints import Controller, param",
+            "class Default(Controller):",
+            "    def ANY(self, **kwargs):",
+            "        return 'any'",
+            "    def POST(self, **kwargs):",
+            "        return 'post'",
+        ])
+
+        res = c.handle("/")
+        self.assertEqual(200, res.code)
+        self.assertEqual("any", res.body)
+
+        res = c.handle("/", method="POST")
+        self.assertEqual(200, res.code)
+        self.assertEqual("post", res.body)
+
+    def test_unsupported_method_404(self):
+        c = Server(contents=[
+            "from endpoints import Controller, param",
+            "class Default(Controller):",
+            "    def POST(self, foo):",
+            "        pass",
+        ])
+        res = c.handle("/foo/bar")
+        self.assertEqual(404, res.code)
+
+        res = c.handle("/foo")
+        self.assertEqual(404, res.code)
+
+        res = c.handle("/")
+        self.assertEqual(501, res.code)
+
     def test_cors(self):
         class Cors(Controller):
             def POST(self): pass
@@ -42,8 +95,7 @@ class ControllerTest(TestCase):
 
     def test_bad_typeerror(self):
         """There is a bug that is making the controller method is throw a 404 when it should throw a 500"""
-        controller_prefix = "badtyperr"
-        c = Server(controller_prefix, [
+        c = Server(contents=[
             "from endpoints import Controller",
             "class Default(Controller):",
             "    def GET(self):",
@@ -52,8 +104,7 @@ class ControllerTest(TestCase):
         res = c.handle('/')
         self.assertEqual(500, res.code)
 
-        controller_prefix = "badtyperr2"
-        c = Server(controller_prefix, [
+        c = Server(contents=[
             "from endpoints import Controller",
             "class Bogus(object):",
             "    def handle_controller(self, foo):",
@@ -324,9 +375,6 @@ class RouterTest(TestCase):
 
         for t in ts:
             req, res = self.get_http_instances(**t['in'])
-#             r = Request()
-#             for key, val in t['in'].items():
-#                 setattr(r, key, val)
 
             r = Router([controller_prefix])
             d = r.find(req, res)
@@ -335,8 +383,109 @@ class RouterTest(TestCase):
 
 
 class CallTest(TestCase):
-    def test_lowercase_method(self):
+    def test_routing_error_unexpected_args(self):
+        c = Server(contents=[
+            "from endpoints import Controller, param",
+            "class Default(Controller):",
+            "    def GET(self, foo):",
+            "        pass",
+        ])
+        res = c.handle("/foo/bar", query_kwargs=dict(foo=2))
+        self.assertEqual(404, res.code)
 
+        c = Server(contents=[
+            "from endpoints import Controller, param",
+            "class Default(Controller):",
+            "    def GET(self, foo):",
+            "        pass",
+        ])
+        res = c.handle("/foo/bar", query_kwargs=dict(che=2))
+        self.assertEqual(404, res.code)
+
+        c = Server(contents=[
+            "from endpoints import Controller, param",
+            "class Default(Controller):",
+            "    def GET(self, foo):",
+            "        pass",
+        ])
+        res = c.handle("/foo/bar", query_kwargs=dict(foo=1, bar=2))
+        self.assertEqual(404, res.code)
+
+        c = Server(contents=[
+            "from endpoints import Controller, param",
+            "class Default(Controller):",
+            "    def GET(self, foo):",
+            "        pass",
+        ])
+        res = c.handle("/foo", query_kwargs=dict(foo=1, bar=2))
+        self.assertEqual(409, res.code)
+
+        c = Server(contents=[
+            "from endpoints import Controller, param",
+            "class Default(Controller):",
+            "    def GET(self, foo):",
+            "        pass",
+        ])
+        res = c.handle("/foo", query_kwargs=dict(bar=2))
+        self.assertEqual(405, res.code)
+
+        c = Server(contents=[
+            "from endpoints import Controller, param",
+            "class Default(Controller):",
+            "    def GET(self, foo):",
+            "        pass",
+        ])
+        res = c.handle("/foo/bar")
+        self.assertEqual(404, res.code)
+
+        c = Server(contents=[
+            "from endpoints import Controller, param",
+            "class Default(Controller):",
+            "    def GET(self, foo, **kwargs):",
+            "        pass",
+        ])
+        res = c.handle("/foo/bar", query_kwargs=dict(foo=1, bar=2))
+        self.assertEqual(404, res.code)
+
+    def test_routing_error_no_args(self):
+        c = Server(contents=[
+            "from endpoints import Controller, param",
+            "class Default(Controller):",
+            "    def GET(self):",
+            "        pass",
+        ])
+        res = c.handle("/foo/bar/che/baz/boom/bam/blah")
+        self.assertEqual(404, res.code)
+
+        c = Server(contents=[
+            "from endpoints import Controller, param",
+            "class Default(Controller):",
+            "    def GET(self, **kwargs):",
+            "        pass",
+        ])
+        res = c.handle("/foo/bar/che/baz/boom/bam/blah", query_kwargs=dict(foo=1, bar=2))
+        self.assertEqual(404, res.code)
+
+        c = Server(contents=[
+            "from endpoints import Controller, param",
+            "class Default(Controller):",
+            "    def GET(self, **kwargs):",
+            "        return kwargs",
+        ])
+        res = c.handle("/", query_kwargs=dict(foo=1, bar=2))
+        self.assertEqual(200, res.code)
+        self.assertTrue("foo" in res._body)
+
+        c = Server(contents=[
+            "from endpoints import Controller, param",
+            "class Default(Controller):",
+            "    def GET(self):",
+            "        return kwargs",
+        ])
+        res = c.handle("/", query_kwargs=dict(foo=1, bar=2))
+        self.assertEqual(405, res.code)
+
+    def test_lowercase_method(self):
         c = Server("controller2", {"foo2": [
             "from endpoints import Controller",
             "class Bar(Controller):",
@@ -360,8 +509,7 @@ class CallTest(TestCase):
 
     def test_handle_404_typeerror_1(self):
         """make sure not having a controller is correctly identified as a 404"""
-        controller_prefix = "h404te"
-        c = Server(controller_prefix, [
+        c = Server(contents=[
             "from endpoints import Controller, Redirect",
             "class NoFoo(Controller):",
             "    def GET(*args, **kwargs):",
@@ -406,11 +554,10 @@ class CallTest(TestCase):
         res = c.handle("/", "POST")
         self.assertEqual(405, res.code)
 
-    def test_handle_405_typeerror_3(self):
+    def test_handle_404_typeerror_3(self):
         """there was an error when there was only one expected argument, turns out
         the call was checking for "arguments" when the message just had "argument" """
-        controller_prefix = "h404te3"
-        c = Server(controller_prefix, [
+        c = Server(contents=[
             "from endpoints import Controller",
             "class Foo(Controller):",
             "    def GET(self): pass",
@@ -418,7 +565,7 @@ class CallTest(TestCase):
         ])
 
         res = c.handle("/foo/bar/baz", query='che=1&boo=2')
-        self.assertEqual(405, res.code)
+        self.assertEqual(404, res.code)
 
     def test_handle_accessdenied(self):
         """raising an AccessDenied error should set code to 401 and the correct header"""
